theory KBO_More
  imports KBO
    Term_Order
    Simple_Termination
begin

text \<open>Incrementality of KBO w.r.t. precedence.\<close>
locale two_kbos = kbo1: kbo w w0 scf least1 S1 W1 + kbo2: kbo w w0 scf least2 S2 W2
  for w w0 scf and least1 :: "'f \<Rightarrow> bool" and S1 W1 least2 S2 W2
begin
lemma kbo_prec_mono: 
  assumes least_imp: "\<And>f::'f. least1 f \<Longrightarrow> least2 f"
    and S_imp: "\<And>fn gm. S1 fn gm \<Longrightarrow> S2 fn gm"
    and W_imp: "\<And>fn gm. W1 fn gm \<Longrightarrow> W2 fn gm"
  shows "(kbo1.S s t \<longrightarrow> kbo2.S s t) \<and> (kbo1.NS s t \<longrightarrow> kbo2.NS s t)"
    (is "(?S s t \<longrightarrow> ?S' s t) \<and> (?N s t \<longrightarrow> ?N' s t)")
proof (induct rule: kbo1.kbo.induct)
  case (1 s t)
  note simps = kbo1.kbo.simps kbo2.kbo.simps
  note [simp del] = simps
  note lex_mono = lex_ext_unbounded_mono[of _ _ "kbo1.kbo" "kbo2.kbo"]
  note lex_mono1 = lex_mono[THEN conjunct1, rule_format]
  note lex_mono2 = lex_mono[THEN conjunct2, rule_format]
  show ?case unfolding simps[of s t] using 1 
    by (auto split: term.splits simp: least_imp S_imp W_imp intro!: lex_mono1 lex_mono2)
qed
end

context admissible_kbo
begin


section \<open>use KBO as ce-compatible reduction pair / triple\<close>

abbreviation kbo_S where "kbo_S \<equiv> {(s,t). S s t}"
abbreviation kbo_NS where "kbo_NS \<equiv> {(s,t). NS s t}"

lemma kbo_NS_supteq: "s \<unrhd> t \<Longrightarrow> (s,t) \<in> kbo_NS"
  using NS_supteq by blast

lemma kbo_ce_compat: "ce_trs cn \<subseteq> kbo_S \<inter> kbo_NS"
proof -
  obtain c n where cn: "cn = (c,n)" by force
  {
    fix s t
    assume "(s,t) \<in> ce_trs cn"
    then have "s \<rhd> t" unfolding cn ce_trs.simps by auto
    from S_supt[OF this] have "(s,t) \<in> kbo_S" by simp
    with S_imp_NS[of s t] have "(s,t) \<in> kbo_S \<inter> kbo_NS" by auto
  }
  then show ?thesis by auto
qed

lemma S_ce_compat: "ce_trs cn \<subseteq> kbo_S" using kbo_ce_compat by blast
lemma NS_ce_compat: "ce_trs cn \<subseteq> kbo_NS" using kbo_ce_compat by blast

lemma kbo_redpair: "mono_ce_af_redtriple_order kbo_S kbo_NS kbo_NS full_af"
proof (unfold_locales)
  show "SN kbo_S" using kbo_strongly_normalizing unfolding SN_defs by blast
  show "subst.closed kbo_S" using S_subst by auto
  show "subst.closed kbo_NS" using NS_subst by auto
  show "ctxt.closed kbo_NS" 
    by (rule, insert NS_ctxt, blast)
  show "kbo_S O kbo_NS \<subseteq> kbo_S" using S_NS_compat by auto
  show "kbo_NS O kbo_S \<subseteq> kbo_S" using NS_S_compat by auto
  show "trans kbo_S" using S_trans unfolding trans_def by auto
  show "trans kbo_NS" using NS_trans unfolding trans_def by auto
  show "refl kbo_NS" using NS_refl unfolding refl_on_def by auto
  show "af_compatible full_af kbo_NS" by (rule full_af)
  show "\<exists> n. \<forall> m \<ge> n. \<forall> c. ce_trs (c,m) \<subseteq> kbo_NS"
    using NS_ce_compat by blast
  show "\<exists> n. \<forall> m \<ge> n. \<forall> c . ce_trs (c,m) \<subseteq> kbo_S"
    using S_ce_compat by blast
  show "kbo_S \<subseteq> kbo_NS" using S_imp_NS by blast
qed

lemma kbo_S_mono: "ctxt.closed kbo_S"
  by (rule, insert S_ctxt, blast)



section \<open>Strong Normalization via Kruskal\<close>

lemma pr_strict_reflclp_imp_pr_weak:
  assumes "pr_strict\<^sup>=\<^sup>= x y" shows "pr_weak x y"
  using assms by (auto simp: pr_strict)

lemma pr_strict_trans:
  "pr_strict fn gm \<Longrightarrow> pr_strict gm hk \<Longrightarrow> pr_strict fn hk"
  using pr_weak_trans
  unfolding pr_strict by blast

lemma transp_on_pr_strict: "transp_on pr_strict UNIV"
  using pr_strict_trans unfolding transp_on_def by blast

lemma irreflp_on_pr_strict: "irreflp_on pr_strict UNIV"
  using pr_strict_irrefl unfolding irreflp_on_def by blast

lemma po_on_pr_strict: "po_on pr_strict\<inverse>\<inverse> UNIV"
proof
  show "irreflp_on pr_strict\<inverse>\<inverse> UNIV"
    by (simp) (rule irreflp_on_pr_strict)
next
  show "transp_on pr_strict\<inverse>\<inverse> UNIV"
    by (simp) (rule transp_on_pr_strict)
qed

lemma wfp_on_pr_strict: "wfp_on pr_strict\<inverse>\<inverse> (UNIV::('f \<times> nat) set)"
  using pr_SN unfolding wfp_on_SN_conv by auto

definition "scfs fn = map (scf fn) [0 ..< snd fn]"

lemma weight_list_conv:
  "sum_list (map (\<lambda>(t\<^sub>i, i). weight t\<^sub>i * scf (f, n) i) (zip ts [0 ..< n])) =
   sum_list (map (\<lambda>(t\<^sub>i, c\<^sub>i). weight t\<^sub>i * c\<^sub>i) (zip ts (scfs (f, n))))"
  by (auto simp: scfs_def zip_map2 o_def intro: arg_cong)

lemma S_subterm_one:
  assumes "t \<in> set ts"
  shows "S (Fun f ts) t"
proof -
  from \<open>t \<in> set ts\<close> have "Fun f ts \<rhd> t" by auto
  with S_supt show ?thesis by blast
qed

text \<open>Lemmas for \emph{simple weights} (i.e., without subterm coefficient functions).\<close>
context
  assumes scf_1: "i < n \<Longrightarrow> scf (f, n) i = 1"
begin

lemma weight_simp' [simp]:
  "weight (Fun f ts) = w (f, length ts) + (\<Sum>t\<leftarrow>ts. weight t)"
proof -
  define n where "n = length ts"
  have *: "zip (map weight ts) (map (scf (f, n)) [0..<n]) = map (\<lambda>t. (weight t, 1)) ts"
    (is "?xs = ?ys")
  proof (rule nth_equalityI)
    show length_eq: "length ?xs = length ?ys" by (simp add: n_def)
    show "\<forall>i<length ?xs. ?xs ! i = ?ys ! i"
    proof (intro allI impI)
      fix i
      assume "i < length ?xs"
      then have *: "i < length (map weight ts)" "i < length (map (scf (f, n)) [0..<n])"
        and "i < n"
        by auto
      then have len': "i < length [0..<n]" by (simp add: nth_def)
      from \<open>i < n\<close> have ith: "[0..<n] ! i = i" by simp
      have map_ith: "map (scf (f, n)) [0..<n] ! i = 1"
        using scf_1 [OF \<open>i < n\<close>] by (simp add: ith nth_map [OF len'])
      show "?xs ! i = ?ys ! i"
        unfolding nth_zip [OF *] map_ith
        using \<open>i < n\<close> unfolding n_def by simp
    qed
  qed
  have "(\<Sum>(ti, i)\<leftarrow>zip ts [0..<n]. weight ti * scf (f, n) i) =
    (\<Sum>tii\<leftarrow>zip ts [0..<n]. weight (fst tii) * scf (f, n) (snd tii))"
    unfolding split_def ..
  also have "\<dots> = (\<Sum>wic\<leftarrow>zip (map weight ts) (map (scf (f, n)) [0..<n]). fst wic * snd wic)"
    by (simp add: zip_map_map split_def o_def)
  also have "\<dots> = (\<Sum>wic\<leftarrow>map (\<lambda>t. (weight t, 1)) ts. fst wic * snd wic)" unfolding * ..
  also have "\<dots> = (\<Sum>wic\<leftarrow>map (\<lambda>t. (weight t, 1::nat)) ts. fst wic * 1)" by (induct ts) auto
  also have "\<dots> = (\<Sum>t\<leftarrow>ts. weight t)" by (induct ts) auto
  finally show ?thesis unfolding weight.simps by (simp add: n_def)
qed

lemma subseq_weight:
  assumes "subseq ss ts"
  shows "(\<Sum>s\<leftarrow>ss. weight s) \<le> (\<Sum>t\<leftarrow>ts. weight t)"
  using assms by (induct) simp+

lemma concat_map_singleton:
  assumes "\<And>x. x \<in> set xs \<Longrightarrow> f x = [g x]"
  shows "concat (map f xs) = map g xs"
  using assms by (induct xs) simp_all

lemma wqo_on_map':
  fixes P and Q and h
  defines "P' \<equiv> \<lambda>x y. P x y \<and> Q (h x) (h y)"
  assumes "wqo_on P\<^sup>=\<^sup>= A"
    and "wqo_on Q B"
    and subset: "h ` A \<subseteq> B"
  shows "wqo_on P'\<^sup>=\<^sup>= A"
proof
  let ?Q = "\<lambda>x y. Q (h x) (h y)"
  from \<open>wqo_on P\<^sup>=\<^sup>= A\<close> have "transp_on P\<^sup>=\<^sup>= A"
    by (rule wqo_on_imp_transp_on)
  then show "transp_on P'\<^sup>=\<^sup>= A"
    using \<open>wqo_on Q B\<close> and subset
    unfolding wqo_on_def transp_on_def P'_def by blast

  from \<open>wqo_on P\<^sup>=\<^sup>= A\<close> have "almost_full_on P\<^sup>=\<^sup>= A"
    by (rule wqo_on_imp_almost_full_on)
  from \<open>wqo_on Q B\<close> have "almost_full_on Q B"
    by (rule wqo_on_imp_almost_full_on)

  show "almost_full_on P'\<^sup>=\<^sup>= A"
  proof
    fix f
    assume *: "\<forall>i::nat. f i \<in> A"
    from almost_full_on_imp_homogeneous_subseq [OF \<open>almost_full_on P\<^sup>=\<^sup>= A\<close> this]
      obtain g :: "nat \<Rightarrow> nat"
      where g: "\<And>i j. i < j \<Longrightarrow> g i < g j"
      and **: "\<forall>i. f (g i) \<in> A \<and> P\<^sup>=\<^sup>= (f (g i)) (f (g (Suc i)))"
        using * by auto
    from chain_transp_on_less [OF ** \<open>transp_on P\<^sup>=\<^sup>= A\<close>]
      have **: "\<And>i j. i < j \<Longrightarrow> P\<^sup>=\<^sup>= (f (g i)) (f (g j))" .
    let ?g = "\<lambda>i. h (f (g i))"
    from * and subset have B: "\<And>i. ?g i \<in> B" by auto
    with \<open>almost_full_on Q B\<close> [unfolded almost_full_on_def good_def, THEN bspec, of ?g]
      obtain i j :: nat
      where "i < j" and "Q (?g i) (?g j)" by blast
    with ** [OF \<open>i < j\<close>] have "P'\<^sup>=\<^sup>= (f (g i)) (f (g j))"
      by (auto simp: P'_def)
    with g [OF \<open>i < j\<close>] show "good P'\<^sup>=\<^sup>= f" by (auto simp: good_def)
  qed
qed

lemma scf_list_id [simp]:
  "scf_list (scf (f, length ss)) ss = ss"
proof -
  let ?scf = "scf (f, length ss)"
  have [simp]: "\<And>i x. i < length ss \<Longrightarrow> replicate (?scf i) x = [x]"
    using scf_1 [of _ "length ss" f] by simp
  have "concat (map (\<lambda>(x, i). replicate (?scf i) x) (zip ss [0 ..< length ss]))
    = map fst (zip ss [0 ..< length ss])"
    by (rule concat_map_singleton) (auto simp: in_set_zip)
  also have "\<dots> = ss" by (auto)
  finally show ?thesis by (simp add: scf_list_def)
qed

lemma SCF_id [simp]:
  "SCF = (\<lambda>x. x)"
proof
  fix t :: "('f, 'v) term"
  show "SCF t = t" by (induct t) (auto simp: map_idI)
qed

lemma subseq_vars_term_ms:
  assumes "subseq ts ss"
  shows "vars_term_ms (SCF (Fun g ts)) \<subseteq># vars_term_ms (SCF (Fun f ss))"
  using assms
proof (induct)
  case (list_emb_Cons xs ys y)
  obtain m1 m2 where id: "\<Union>#image_mset vars_term_ms (mset xs) = m1" 
    "\<Union>#image_mset vars_term_ms (mset ys) = m2" by auto
  from list_emb_Cons
  have IH: "m1 \<subseteq># m2" by (auto simp: id)
  also have "\<dots> \<subseteq># vars_term_ms y + m2" by (rule mset_subset_eq_add_right)
  finally show ?case by (simp add: id)
qed auto  

text \<open>
  Without subterm coefficient functions, @{term S} is a simplification-order.
\<close>
lemma simplification_order:
  assumes "wqo_on pr_strict\<inverse>\<inverse>\<^sup>=\<^sup>= UNIV"
  shows "simplification_order {(x :: ('f, 'v) term, y). S x y}"
  (is "simplification_order ?kbo")
proof (unfold simplification_order_def)
  let ?P = "\<lambda>x y. pr_strict x y \<and> w x \<ge> w y"
  have "range w \<subseteq> UNIV" by auto
  from wqo_on_map' [OF assms wqo_on_UNIV_nat this]
    have "wqo_on ?P\<inverse>\<inverse>\<^sup>=\<^sup>= UNIV" by (auto simp: conversep_iff [abs_def])
  moreover
  have "emb UNIV ?P \<subseteq> ?kbo"
  proof (rule emb_subsetI)
    show "\<And>s t u. (s, t) \<in> ?kbo \<Longrightarrow> (t, u) \<in> ?kbo \<Longrightarrow> (s, u) \<in> ?kbo"
      and "\<And>C s t. (s, t) \<in> ?kbo \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> ?kbo"
      and "\<And>\<sigma> s t. (s, t) \<in> ?kbo \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ?kbo"
      by (auto simp del: kbo.simps dest: S_trans S_ctxt S_subst)
  next
    fix t f and ts :: "('f, 'v) term list"
    assume "t \<in> set ts"
    from S_subterm_one [OF this]
      show "(Fun f ts, t) \<in> ?kbo" by simp
  next
    fix f g and ss ts :: "('f, 'v) term list"
    presume strict: "pr_strict (f, length ss) (g, length ts)"
      and w: "w (f, length ss) \<ge> w (g, length ts)"
      and "subseq ts ss"
    have "vars_term_ms (SCF (Fun g ts)) \<subseteq># vars_term_ms (SCF (Fun f ss))"
      by (rule subseq_vars_term_ms) fact
    moreover
    have "weight (Fun f ss) \<ge> weight (Fun g ts)"
      using subseq_weight [OF \<open>subseq ts ss\<close>] and w
      unfolding weight_simp' by simp
    ultimately
    show "(Fun f ss, Fun g ts) \<in> ?kbo" using strict by (simp add: kbo.simps)
  qed simp_all
  moreover
  have "rewrite_order ?kbo"
    by (auto simp del: kbo.simps
             simp: rewrite_order_def rewrite_relation_def irrefl_def S_irrefl trans_def
             dest: S_ctxt S_subst S_trans)
  ultimately
  show "rewrite_order ?kbo \<and> (\<exists>Q. wqo_on Q\<inverse>\<inverse>\<^sup>=\<^sup>= UNIV \<and> emb UNIV Q \<subseteq> ?kbo)" by blast
qed

text \<open>
  Alternative termination proof for the case that @{term "pr_strict\<inverse>\<inverse>\<^sup>=\<^sup>="} is a
  well-quasi-order.
\<close>
lemma 
  assumes "wqo_on pr_strict\<inverse>\<inverse>\<^sup>=\<^sup>= UNIV"
  shows "SN {(x :: ('f, 'v) term, y). S x y}"
  by (rule simplification_order_imp_SN [OF simplification_order [OF assms]])

declare weight_simp' [simp del]
end
end

end
